// notify.cpp

#include "notify.h"

#include <print>
#include <ranges>
#include <stdexcept>
#include <string>
#include <utility>

#include "libnotify_types.h"

namespace notify {

namespace rv = std::ranges::views;
using std::operator""sv;

namespace {

[[nodiscard]] constexpr auto icon_to_sv(DialogIcon icon) -> std::string_view {
	switch (icon) {
		case DialogIcon::Error:
			return "dialog-error"sv;
		case DialogIcon::Warning:
			return "dialog-warning"sv;
		case DialogIcon::Info:
			return "dialog-information"sv;
		case DialogIcon::Question:
			return "dialog-question"sv;
		case DialogIcon::Password:
			return "dialog-password"sv;
	}
	std::unreachable();
}

}  // namespace

struct NotifyInterface {
	NotifyInterface() = default;
	virtual ~NotifyInterface() = default;

	NotifyInterface(NotifyInterface const&) = delete;
	auto operator=(NotifyInterface const&) -> NotifyInterface& = delete;

	NotifyInterface(NotifyInterface&&) = default;
	auto operator=(NotifyInterface&&) -> NotifyInterface& = default;

	virtual auto show(std::string_view title, std::string_view message,
					  DialogIcon icon) -> bool = 0;
	[[nodiscard]] virtual auto error_message() const -> std::string_view = 0;
};

class LibnotifyBackend final : public NotifyInterface {
   public:
	LibnotifyBackend(LibnotifyBackend const&) = delete;
	LibnotifyBackend(LibnotifyBackend&&) = delete;
	auto operator=(LibnotifyBackend const&) -> LibnotifyBackend& = delete;
	auto operator=(LibnotifyBackend&&) -> LibnotifyBackend& = delete;

	explicit LibnotifyBackend(std::string_view app_name) : last_error_{} {
		if (!(static_cast<bool>(libnotify::initialized()) ||
			  static_cast<bool>(libnotify::init(app_name.data())))) {
			throw std::runtime_error("Failed to initialize libnotify");
		}
	}

	~LibnotifyBackend() override {
		if (static_cast<bool>(libnotify::initialized())) {
			libnotify::destroy();
		}
	}

	auto show(std::string_view title, std::string_view message, DialogIcon icon) -> bool override {
		libnotify::NotifyNotification* notification =
			libnotify::notification_new(title.data(), message.data(), icon_to_sv(icon).data());
		if (notification == nullptr) {
			last_error_ = "Failed to create notification";
			return false;
		}

		libnotify::GError* error = nullptr;
		auto result = static_cast<bool>(libnotify::notification_show(notification, &error));
		if (!result) {
			last_error_ = (error != nullptr) ? error->message : "Unknown error";
			libnotify::free_error(error);
		}

		libnotify::free_g_obj(G_OBJECT(notification));
		return result;
	}

	[[nodiscard]] auto error_message() const -> std::string_view override { return last_error_; }

   private:
	std::string last_error_;
};

class ConsoleBackend final : public NotifyInterface {
   public:
	explicit ConsoleBackend(std::string_view app_name) : app_name_(app_name) {}

	auto show(std::string_view title, std::string_view message, DialogIcon icon) -> bool override {
		auto notification_type = icon_to_sv(icon) | rv::drop("dialog-"sv.size()) |
								 rv::transform(::toupper) | std::ranges::to<std::string>();
		std::print(stderr, "[{}] {} - {}: {}\n", app_name_, notification_type, title, message);
		return true;
	}

	[[nodiscard]] auto error_message() const -> std::string_view override { return {}; }

   private:
	std::string app_name_;
};

template <typename T>
inline constexpr auto always_false_v = false;

template <BackendType Backend>
Notification<Backend>::Notification(std::string_view app_name) : impl_{} {
	if constexpr (Backend == BackendType::Libnotify) {
		impl_ = std::make_unique<LibnotifyBackend>(app_name);
	} else if constexpr (Backend == BackendType::Console) {
		impl_ = std::make_unique<ConsoleBackend>(app_name);
	} else {
		static_assert(always_false_v<decltype(Backend)>, "Backend not implemented");
	}
}

template <BackendType Backend>
Notification<Backend>::~Notification() = default;

template <BackendType Backend>
Notification<Backend>::Notification(Notification&&) noexcept = default;

template <BackendType Backend>
auto Notification<Backend>::operator=(Notification&&) noexcept -> Notification& = default;

template <BackendType Backend>
auto Notification<Backend>::show(std::string_view title, std::string_view message,
								 DialogIcon icon) -> bool {
	return impl_->show(title, message, icon);
}

template <BackendType Backend>
auto Notification<Backend>::error_message() const -> std::string_view {
	return impl_->error_message();
}

// Explicit instantiations, let LTO take care of it
template class Notification<BackendType::Libnotify>;
template class Notification<BackendType::Console>;

}  // namespace notify
