// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright © 2023 Thorsten Schubert <tschubert@bafh.org>

#include <re2/re2.h>

#include <CLI/CLI.hpp>
#include <CLI/Error.hpp>
#include <algorithm>
#include <charconv>
#include <chrono>
#include <concepts>
#include <csignal>
#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstdlib>
#include <deque>
#include <filesystem>
#include <limits>
#include <print>
#include <ranges>
#include <span>
#include <string>
#include <string_view>
#include <thread>
#include <unordered_map>
#include <unordered_set>
#include <utility>

#include "config.h"
#include "debug.h"
#include "notify.h"

#if __has_include(<generator>)
#include <generator>
#endif

#if __has_include(<experimental/simd>)
#include <experimental/simd>
#endif

#if defined(__cpp_lib_generator) && __cpp_lib_generator >= 202207L
namespace pw {
template <typename R, typename V = void, typename A = void>
using generator = std::generator<R, V, A>;
}
#else
// NOLINTNEXTLINE
#warning "using generator compatibility layer"
#include "compat/generator.hpp"
#endif

namespace {

namespace rn = std::ranges;
namespace rv = std::ranges::views;

using std::operator""sv;

using i64 = std::int_fast64_t;
using u64 = std::uint_fast64_t;

using ProcessId = u64;
using ProccessName = std::string;
using ProcessCmdline = std::string;

using TimeMs = u64;
using CpuJiffies = u64;
using UtimeJiffies = CpuJiffies;
using StimeJiffies = CpuJiffies;
using CpuTimePair = std::pair<UtimeJiffies, StimeJiffies>;

using Pattern = std::string;
using PatternList = std::vector<Pattern>;

struct ProcessCmd {
	ProcessId pid;
	ProcessCmdline name;
};

using ProcessCmdList = std::vector<ProcessCmd>;

template <typename To, typename From>
__attribute__((always_inline)) constexpr auto narrow_cast(From&& u) noexcept -> To {
	return static_cast<To>(std::forward<From>(u));
}

struct AppConfig {
	PatternList pattern;
	u64 cycles;
	u64 interval;
	u64 grace_period;
	i64 nproc;
	i64 tps;
	float threshold;
};

struct FileError : std::runtime_error {
	explicit FileError(std::string const& message, std::string const& data = {})
		: std::runtime_error(data.empty() ? message : std::format("{}: {}"sv, message, data)) {}
};

struct FileNotFoundError final : FileError {
	explicit FileNotFoundError(std::string const& message, std::string const& data = {})
		: FileError(message, data) {}
};

struct FileParseError final : FileError {
	explicit FileParseError(std::string const& message, std::string const& data = {})
		: FileError(message, data) {}
};

struct ProcessExited final : FileError {
	explicit ProcessExited(u64 pid)
		: FileError(std::format("process with PID {} already exited"sv, pid)) {}
};

}  // namespace

// Enables formatted output of AppConfig using std::format
template <>
struct std::formatter<AppConfig> {
	template <typename... Args>
	static consteval auto total_digits10() -> std::size_t {
		return (std::numeric_limits<Args>::digits10 + ...);
	}

	static constexpr auto parse(std::format_parse_context& ctx) { return ctx.begin(); }

	static auto format(AppConfig const& cfg, format_context& ctx) {
		auto out = ctx.out();

		// Write initial opening
		out = std::format_to(out, "AppConfig {{\n  pattern: ");

		// Pattern list with exact same formatting as your version
		out = std::format_to(out, "{{\n");
		for (auto const& pattern : cfg.pattern) {
			out = std::copy_n("    ", 4, out);
			out = rn::copy(pattern, out).out;
			out = std::copy_n(",\n", 2, out);
		}
		out = std::copy_n("  }", 3, out);

		// Rest identical to your version
		return std::format_to(out,
							  ",\n"
							  "  cycles: {},\n"
							  "  threshold: {},\n"
							  "  interval: {},\n"
							  "  grace_period: {},\n"
							  "  nproc: {},\n"
							  "  tps: {}\n"
							  "}};"sv,
							  cfg.cycles, cfg.threshold, cfg.interval, cfg.grace_period, cfg.nproc,
							  cfg.tps);
	}
};

namespace {

// Reads data from a file specified by 'path' into a pre-allocated contiguous memory buffer 'buf'
// and returns the number of bytes read, which is at most 'size'
auto bounded_file_read(std::string_view path, std::span<char> buf) -> std::size_t {
	std::unique_ptr<std::FILE, decltype([](FILE* fd) noexcept { std::fclose(fd); })> file(
		std::fopen(path.data(), "r"));

	if (buf.size() == 0ZU) [[unlikely]] {
		return 0ZU;
	}

	if (!file) {
		throw FileNotFoundError("failed to open file", std::string{path});
	}

	auto total{0ZU};

	// clang-format off
	do [[unlikely]] {
		auto const bytes =
			std::fread(buf.data() + total, sizeof(char), buf.size() - total, file.get());
		if (bytes == 0) {
			if (std::ferror(file.get()) != 0) [[unlikely]] {
				throw FileError("error reading file", std::string{path});
			}
			break;	// EOF
		}
		total += bytes;
	} while (total < buf.size());
	// clang-format on

	return total;
}

// Yields process command lines as a generator
[[nodiscard]] auto yield_process_cmdline() -> pw::generator<ProcessCmd> {
	static constexpr auto CMD_SIZE = 1ZU << 10;
	static constexpr auto PATH_SIZE = 1ZU << 8;

	std::string path{};
	std::array<char, CMD_SIZE> cmd{};
	path.reserve(PATH_SIZE);

	for (auto const& entry : std::filesystem::directory_iterator{
			 "/proc"sv, std::filesystem::directory_options::skip_permission_denied}) {
		if (!entry.is_directory())
			continue;

		path.assign(entry.path().filename());
		// Reject all files with non-digits inside the name
		if (!rn::all_of(path, static_cast<int (*)(int)>(std::isdigit))) {
			continue;
		}

		std::string_view pv{path};
		u64 pid{};
		if (auto const [_, ec] = std::from_chars(pv.data(), pv.data() + pv.size(), pid);
			ec != std::errc()) {
			continue;
		}

		auto cmd_path = entry.path() / "cmdline"sv;
		try {
			if (bounded_file_read(cmd_path.c_str(), cmd) == 0ZU) {
				// e.g. kworkers
				continue;
			}
		} catch (FileNotFoundError const&) {
			// This always means the process exited before the read was successful which is a
			// possibility for short-lived processes
			DEBUG_PRINTF("Process with PID {} disappeared during process enumeration\n"sv, path);
			continue;
		} catch (FileError const& e) {
			std::print(stderr, "{}\n"sv, e.what());
			continue;
		}

		// yield only the null-terminate command name field
		co_yield {.pid = pid, .name = std::string{cmd.data()}};
	}
}

// Parses CPU time for a given process ID
[[nodiscard]] auto parse_cpu_time(u64 pid) -> CpuTimePair {
	static constexpr auto STAT_SIZE = 1ZU << 10;
	std::array<char, STAT_SIZE> sbuf{};
	std::array<char, std::numeric_limits<i64>::digits10 + /*\0*/ 1ZU> pbuf{};

	auto [_, ec] = std::to_chars(pbuf.data(), pbuf.data() + pbuf.size(), pid);
	if (ec != std::errc()) [[unlikely]] {
		throw FileParseError{"PID conversion error", pbuf.data()};
	}

	auto const path = std::filesystem::path("/proc"sv) / pbuf.data() / "stat"sv;
	try {
		if (bounded_file_read(path.c_str(), sbuf) == 0ZU) [[unlikely]] {
			throw FileParseError{"no data read", path.c_str()};
		}
	} catch (FileNotFoundError const&) {
		// This time we actually care about process disappearance since we are trying to parse the
		// cpu time fields and it is potentially viewed as active
		DEBUG_PRINTF("Process with PID {} disappeared during cpu time parsing\n"sv, pbuf.data());
		throw ProcessExited{pid};
	}

	std::string_view line{sbuf.data()};
	auto const comm = line.find_last_of(')');
	if (comm == std::string::npos || comm + 2 > line.length()) [[unlikely]] {
		throw FileParseError{"comm field parse error", path.c_str()};
	}

	auto const proc_stats = line.substr(comm + 2);
	auto tokens = proc_stats | rv::split(' ') | rv::transform([](auto const& range) noexcept {
					  return std::string_view{&*range.cbegin(),
											  static_cast<std::size_t>(rn::distance(range))};
				  });

	static constexpr auto utime_pos = 11;
	u64 utime{}, stime{};
	auto it = tokens.cbegin();

	// utime
	if (it = std::next(it, utime_pos); it == tokens.end()) [[unlikely]] {
		throw FileParseError{"utime parse error", path.c_str()};
	}

	auto token = *it;
	if (auto const [_, ec] = std::from_chars(token.data(), token.data() + token.size(), utime);
		ec != std::errc()) [[unlikely]] {
		throw FileParseError{"utime conversion error", path.c_str()};
	}

	// stime
	if (++it == tokens.end()) [[unlikely]] {
		throw FileParseError{"stime parse error", path.c_str()};
	}

	token = *it;
	if (auto const [_, ec] = std::from_chars(token.data(), token.data() + token.size(), stime);
		ec != std::errc()) [[unlikely]] {
		throw FileParseError{"stime conversion error", path.c_str()};
	}

	return std::make_pair(utime, stime);
}

template <std::floating_point Float>
constexpr auto checked_mul(Float a, Float b) -> std::optional<Float> {
	// NOLINTNEXTLINE
	[[assume(!std::isnan(a) && !std::isnan(b))]];

#ifdef __cpp_lib_experimental_parallel_simd
	namespace stdx = std::experimental;
	if constexpr (stdx::is_simd_v<stdx::native_simd<Float>>) {
		using simd_t = stdx::native_simd<Float>;
		simd_t const va(a);
		simd_t const vb(b);
		simd_t const vresult = va * vb;

		auto const is_inf = stdx::isinf(vresult);
		if (is_inf[0]) {
			return std::nullopt;
		}

		return vresult[0];
	} else
#endif
	{
		Float const result = a * b;

		if (std::isinf(result)) {
			return std::nullopt;
		}

		constexpr Float epsilon = std::numeric_limits<Float>::epsilon();
		constexpr Float min_normal = std::numeric_limits<Float>::min();
		if (std::abs(result) < min_normal && std::abs(a) > epsilon && std::abs(b) > epsilon) {
			return std::nullopt;
		}

		return result;
	}
}

// Overflow-checked calculation of total CPU time from user and system time in milliseconds
// static_cast<u64>((narrow_cast<float>(utime) / tps + narrow_cast<float>(stime) / tps) * 1e3F);
__attribute__((const)) auto cpu_time_total(CpuTimePair time, i64 ticks_per_second) noexcept
	-> std::optional<u64> {
	// NOLINTNEXTLINE
	[[assume(ticks_per_second > 0)]];
	static constexpr float ms{1e3F};

	auto const tps = narrow_cast<float>(ticks_per_second);
	// times in jiffies
	auto const [utime, stime] = time;

	return checked_mul(narrow_cast<float>(utime) / tps, ms)
		.and_then([=](float utime_ms) -> std::optional<float> {
			return checked_mul(narrow_cast<float>(stime) / tps, ms)
				.transform([utime_ms](float stime_ms) { return utime_ms + stime_ms; });
		})
		.and_then([=](float total_ms) -> std::optional<u64> {
			if (total_ms <= narrow_cast<float>(std::numeric_limits<u64>::max())) {
				return static_cast<u64>(total_ms);
			}
			return std::nullopt;
		});
}

// Multiple patterns allow for duplicated processes this removes duplicate processes from the list
auto dedup_process_list(ProcessCmdList& procs) -> std::size_t {
	if (procs.empty()) {
		return {};
	}

	auto const size = procs.size();
	rn::sort(procs, rn::less{}, &ProcessCmd::pid);
	auto const [first, last] =
		rn::unique(procs, [](auto const& a, auto const& b) noexcept { return a.pid == b.pid; });
	procs.erase(first, last);

	return size - procs.size();
}

// Test all provided regex patterns against process list and deduplicate results
// NOTE: generators can only be iterated through once, a range does not work here
template <rn::constant_range Range>
[[nodiscard]] auto filter_processes(Range regex_view) -> ProcessCmdList {
	// Exclude own process context
	auto const self = static_cast<ProcessId>(::getpid());

	auto procs =
		regex_view | rv::transform([self](RE2 const& re) {
			return yield_process_cmdline() | rv::filter([&re, self](auto const& proc) noexcept {
					   return proc.pid != self && RE2::PartialMatch(proc.name, re);
				   });
		}) |
		rv::join | rv::as_rvalue | rn::to<ProcessCmdList>();

	[[maybe_unused]] auto dups = dedup_process_list(procs);
	DEBUG_PRINTF("Duplicates removed: {}\n"sv, dups);

	return procs;
}

// Truncates a string with ellipsis if it exceeds the specified length
[[nodiscard]] auto ellipsize(std::string_view buf, std::size_t len = 48ZU,
							 std::string_view suffix = "..."sv) -> std::string {
	if (buf.size() <= len) {
		return std::string{buf};
	}

	if (len <= suffix.size()) [[unlikely]] {
		throw std::invalid_argument("length less than suffix size");
	}

	std::string trunc{};
	trunc.reserve(len);
	trunc.assign(buf.data(), len - suffix.size());
	trunc.append(suffix.data());

	return trunc;
}

// Prints information about matching processes
auto print_process_view(std::span<ProcessCmd> proc_view) -> std::size_t {
	if (proc_view.empty()) {
		return {};
	}

	std::print(stderr, "Matching processes: {}\n{:=<68}\n"sv, proc_view.size(), "="sv);
	rn::for_each(proc_view, [](auto const& proc) noexcept {
		auto const& [pid, name] = proc;
		// Split name by space before ellipsizing
		auto split = name | rv::take_while(std::not_fn(static_cast<int (*)(int)>(std::isspace)));
		std::string_view token{&*split.begin(), static_cast<std::size_t>(rn::distance(split))};
		std::print(stderr, "pid: {:>8} proc: {}\n"sv, pid, ellipsize(token));
	});
	std::print(stderr, "{:=<68}\n"sv, "="sv);

	return proc_view.size();
}

// Sends a notification about high CPU usage
template <notify::BackendType Backend>
auto dispatch_notification(notify::Notification<Backend>& notifier, ProcessCmd const& proc,
						   float cpu) -> void {
	auto const message = std::format("Process with PID {} ({}) is using {:.2f}% CPU."sv, proc.pid,
									 ellipsize(proc.name), cpu);
	std::print(stderr, "{}\n"sv, message);

	// org.freedesktop.Notifications
	if (!notifier.show("High CPU Usage"sv, message, notify::DialogIcon::Warning)) {
		std::print(stderr, "Failed to send notification: \n"sv, notifier.error_message());
	}
}

// Parses command-line arguments and returns application configuration
[[nodiscard]] auto parse_args(CLI::App& app, std::span<char*> args) -> AppConfig {
	AppConfig config{};

	auto const nproc = ::sysconf(_SC_NPROCESSORS_ONLN);
	auto const tps = ::sysconf(_SC_CLK_TCK);

	if (nproc < 1L || tps < 1L) [[unlikely]] {
		throw CLI::ParseError("unable to get system configuration", CLI::ExitCodes::ConfigError);
	}

	config.tps = tps;
	config.nproc = nproc;

	std::map<std::string, uint64_t> const time_units_ms{
		{"ms", 1LU}, {"s", 1'000LU}, {"m", 60'000LU}, {"h", 3'600'000LU}};

	auto const find_largest_prefix = [&time_units_ms](u64 value) noexcept {
		std::vector<decltype(time_units_ms)::const_iterator> iters;

		iters.reserve(time_units_ms.size());
		for (auto it = time_units_ms.begin(); it != time_units_ms.end(); ++it) {
			iters.push_back(it);
		}
		rn::sort(iters, rn::greater{}, [](auto const& it) noexcept { return it->second; });
		for (auto const& it : iters) {
			if (value % it->second == uint64_t{}) {
				return std::to_string(value / it->second).append(it->first);
			}
		}
		return std::to_string(value).append("ms"sv);
	};

	app.description("Monitor CPU usage of processes matching a regular expression pattern");
	app.add_option("-i,--interval", config.interval,
				   "Interval between process cpu usage measurements")
		->transform(CLI::AsNumberWithUnit(time_units_ms))
		->default_val(30'000LU)
		->default_str(find_largest_prefix(config.interval));
	app.add_option("-g,--grace-period", config.grace_period,
				   "Cooldown period before another notification can be sent")
		->transform(CLI::AsNumberWithUnit(time_units_ms))
		->default_val(300'000LU)
		->default_str(find_largest_prefix(config.grace_period));
	app.add_option("-t,--threshold", config.threshold, "CPU usage threshold")
		->check([=](std::string const& str) noexcept -> std::string {
			auto const value = std::stof(str);
			if (value <= 0.0F) {
				return "Threshold must be positive";
			}
			// Assuming NPROC is accessible here
			auto const max = nproc * 100L;
			if (value > narrow_cast<float>(max)) {
				return std::format("Threshold must be ≤ {} (nproc * 100)", max);
			}
			return {};
		})
		->default_val(66.0F)
		->capture_default_str();
	app.add_option(
		   "-c,--cycles", config.cycles,
		   "Consecutive intervals a process must exceed the CPU threshold before notification")
		->default_val(2LU)
		->check(CLI::PositiveNumber);

	app.add_option("<pattern>", config.pattern,
				   "Regular expression pattern to match (partial) process name(s)")
		->required();
	app.set_version_flag("-v,--version", PROJECT_VERSION);

	app.parse(narrow_cast<int>(args.size()), args.data());

	static constexpr auto min_period_ms = 1000LU;
	config.cycles = std::max(config.cycles, 1LU);
	config.interval = std::max(config.interval, min_period_ms);
	config.grace_period = std::max(config.grace_period, min_period_ms);

	return config;
}
}  // namespace

extern "C" {

std::sig_atomic_t volatile g_signal_hup_flag = 0;

// Handles SIGHUP signal
auto signal_handler(int signal) -> void {
	if (signal == SIGHUP) {
		g_signal_hup_flag = 1;
	}
}

}  // extern "C"

// Entry point
auto main(int argc, char** argv) -> int {
	auto const args = std::span{argv, static_cast<std::size_t>(argc)};

	// Parse CLI arguments
	AppConfig cfg{};
	CLI::App app{};
	try {
		cfg = parse_args(app, args);
	} catch (CLI::ParseError const& e) {
		return app.exit(e);
	}
	std::print(stderr, "{}\n", cfg);

	// We do not want RE2/DFA errors printed to stderr.
	RE2::Options options{};
	options.set_log_errors(false);

	std::deque<RE2> re2_list{};
	for (auto const& pattern : cfg.pattern) {
		if (auto const& re = re2_list.emplace_back(pattern, options); !re.ok()) {
			std::print(stderr, "RE2 compilation failed: {}\n"sv, re.error());
			return static_cast<int>(CLI::ExitCodes::InvalidError);
		}
	}
	re2_list.shrink_to_fit();

	auto const rv = rv::all(std::as_const(re2_list));
	auto const print_matching = [&rv]() noexcept {
		// Print a snapshot of matching processes on program start
		auto proc_view = filter_processes(rv);
		print_process_view(proc_view);
	};
	print_matching();

	// SIGHUP -> print matching processes
	std::signal(SIGHUP, signal_handler);

	std::string app_name{std::filesystem::path(args[0]).filename()};

#ifdef NDEBUG
	notify::Notification<notify::BackendType::Libnotify> notification{app_name};
#else
	notify::Notification<notify::BackendType::Console> notification{app_name};
#endif

	static constexpr auto PMAP_SIZE = 1ZU << 10;
	static constexpr auto SLEEP_DURATION = std::chrono::milliseconds{100L};

	auto start = std::chrono::steady_clock::now();
	auto notify_last = start;

	static auto const interval = std::chrono::milliseconds{cfg.interval};
	static auto const grace_period = std::chrono::milliseconds{cfg.grace_period};

	std::unordered_map<ProcessId, TimeMs> prev_cpu_total{}, current_cpu_times{};
	std::unordered_map<ProcessId, u64> over_threshold_cycles{};
	std::unordered_set<ProcessId> active_processes{};

	prev_cpu_total.reserve(PMAP_SIZE), current_cpu_times.reserve(PMAP_SIZE);

	// Main loop, blocks for a fixed duration each iteration
	for (;;) {
		if (g_signal_hup_flag != 0) {
			g_signal_hup_flag = 0;
			DEBUG_PRINTF("{} signal received\n"sv, "SIGHUP"sv);
			std::print(stderr, "{}\n", cfg);
			print_matching();
		}

		// Reuse set to avoid unnecessary allocation
		active_processes.clear();

		auto const now = std::chrono::steady_clock::now();

		// If user-specified interval not yet reached, skip processing
		if (now - start < interval) {
			std::this_thread::sleep_for(SLEEP_DURATION);
			continue;
		}

		// Iterate through matching processes and extract cpu time if interval reached
		start = now;

		auto const filtered = filter_processes(rv);

		{
			DO_SCOPED_MEASUREMENT(__func__);
			for (auto const& proc : filtered) {
				CpuTimePair parsed{};
				try {
					parsed = parse_cpu_time(proc.pid);
				} catch (FileError const& e) {
					std::print(stderr, "{}\n"sv, e.what());
					continue;
				}

				// Track process as active on successful parsing
				active_processes.insert(proc.pid);

				auto const current_time_ms = cpu_time_total(parsed, cfg.tps).value_or(u64{});
				// Save cpu time for next iteration
				current_cpu_times[proc.pid] = current_time_ms;

				// Check if process existed last iteration, if not skip to next entry
				auto const it = prev_cpu_total.find(proc.pid);
				if (it == prev_cpu_total.end()) {
					continue;
				}

				auto const prev_time_ms = it->second;
				// Calculate time delta and convert to percentage
				auto const cpu_percent = narrow_cast<float>(current_time_ms - prev_time_ms) /
										 narrow_cast<float>(cfg.interval) * 100.0F;
				DEBUG_PRINTF("prev: {:>16} current: {:>16} -> {:>.2f} name: {:.48}\n"sv,
							 prev_time_ms, current_time_ms, cpu_percent, proc.name);

				// NOLINTNEXTLINE
				[[assume(cpu_percent >= 0.0f)]];

				// Set threshold not reached -> skip
				if (cpu_percent < cfg.threshold) {
					if (over_threshold_cycles.erase(proc.pid) != 0ZU) {
						std::print(stderr, "Cycle count for PID {} reset\n"sv, proc.pid);
					}
					continue;
				}

				auto& cycles = ++over_threshold_cycles[proc.pid];
				std::print(stderr, "Cycle count for PID {} is at [{}/{}]\n"sv, proc.pid, cycles,
						   cfg.cycles);

				if (cycles < cfg.cycles) {
					continue;
				}

				// Reset cycle count before dispatching a notification
				cycles = 0LU;

				if (now - notify_last < grace_period) {
					auto const remaining = std::chrono::duration_cast<std::chrono::milliseconds>(
						grace_period - (now - notify_last));
					std::print(
						stderr,
						"Grace period active: [{}/{}] - no notification dispatched for PID {} [{:.2f}%]\n"sv,
						remaining, grace_period, proc.pid, cpu_percent);
					continue;
				}

				dispatch_notification(notification, proc, cpu_percent);
				notify_last = now;
			}
		}

		// Erase process from cycle tracking map on disappearance
		std::erase_if(over_threshold_cycles, [&active_processes](auto const& entry) {
			auto inactive = active_processes.find(entry.first) == active_processes.end();
			if (inactive) {
				DEBUG_PRINTF("PID {} disappeared\n"sv, entry.first);
			}
			return inactive;
		});
		DEBUG_PRINTF("Elements in cycle map: {}\n"sv, over_threshold_cycles.size());

		std::swap(prev_cpu_total, current_cpu_times);
		current_cpu_times.clear();
	}
}
